// Copyright (C) 2013 by FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of demoHeterogeneousParallelism.
// 
// demoHeterogeneousParallelism is free software: you can redistribute it and/or modify it under the terms of the GNU 
// General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your 
// option) any later version.
// 
// demoHeterogeneousParallelism is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
// even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
// 
// You should have received a copy of the GNU General Public License along with demoHeterogeneousParallelism. If not, 
// see <http://www.gnu.org/licenses/>.


// "libHawaii" processing steps to wrap existing code: GEMM sample application
// ===========================================================================

#pragma once

#include "hawaii/flow/item.h"
#include "hawaii/flow/step.h"
#include "hawaii/flow/stepGPU.h"
#include "hawaii/GPU/autoMat.h"
#include "hawaii/common/partitionize.h"
#include "GEMM.h"
#include "ompWrapper.h"

// helpers only used here
namespace {
	
	// expand a region of interest back to full size
	void expand( const cv::Mat& ROI,
	                   cv::Mat& whole ) {
		whole = ROI ;
		cv::Size  size   ;
		cv::Point offset ;
		whole.locateROI( size, offset) ;
		whole.adjustROI( offset.y, size.height - offset.y - ROI.rows,
		                 offset.x, size.width  - offset.x - ROI.cols ) ;
	}		
}

// work item for generalized matrix multiplication (GEMM)
class ItemGEMM : public hawaii::flow::ItemTemplate< ItemGEMM > {
	
	// input and result data
	public:
	hawaii::AutoMat A, B, C, // inputs
	                AB,      // intermediate result: "A * B" (used in pipelined GEMM only, empty otherwise)
	                result ; // final result: "A * B * alpha + C * beta"
	
	// require running number to be set
	public:
	ItemGEMM( const uint64_t serialNumberArg,
	          const double   ratioArg        = 1.0 ) :
		ItemTemplate< ItemGEMM >( serialNumberArg, ratioArg ) {
	}
	
	// split work item (required for partitioning only)
	// user note: This assumes that all members have correctly been pre-allocated e.g. by "SourceGEMM"!
	public:
	virtual void splitImpl(       ItemGEMM::PtrVec&           parts,
	                        const std::vector< double >&      ratios,
	                        const hawaii::flow::PartInfoBase& partInfo = hawaii::flow::PartInfoBase() ) override {
		// not needed
		(void)partInfo ;
		
		// partition A, C, ABa and result, but keep whole B
		std::vector< hawaii::AutoMat > AParts, CParts, ABParts, resultParts ;
		AParts      = hawaii::partitionizeHoriz( this->A,      ratios ) ;
		CParts      = hawaii::partitionizeHoriz( this->C,      ratios ) ;
		ABParts     = hawaii::partitionizeHoriz( this->AB,     ratios ) ;
		resultParts = hawaii::partitionizeHoriz( this->result, ratios ) ;
		
		// create partitioned work items
		parts.clear() ;
		for( size_t idx = 0 ; idx < ratios.size() ; ++idx ) {
			ItemGEMM::Ptr part( new ItemGEMM( this->serialNumber, ratios[ idx ] * this->ratio ) ) ;
			part->A      = AParts[      idx ] ;
			part->B      = this->B            ;
			part->C      = CParts[      idx ] ;
			part->AB     = ABParts[     idx ] ;
			part->result = resultParts[ idx ] ;
			parts.push_back( part ) ;
		}
	}
	
	// re-combine work item (required for partitioning only)
	// user note: This assumes that all members have correctly been pre-allocated e.g. by "SourceGEMM"!
	public:
	virtual void mergeImpl( const PtrVec&                     parts,
	                              Ptr&                        whole,
	                        const hawaii::flow::PartInfoBase& partInfo = hawaii::flow::PartInfoBase() ) override {
		// not needed
		(void)partInfo ;
		
		// download all partitions to CPU memory, sum up their ratios
		double ratioWhole = 0.0 ;
		for( const auto& part : parts ) {
			part->A.readCPU()      ;
			part->B.readCPU()      ;
			part->C.readCPU()      ;
			part->AB.readCPU()     ;
			part->result.readCPU() ;
			ratioWhole += part->ratio ;
		}
		
		// allocate whole item: "ratioWhole" is not necessarily 1.0, e.g. with nested partitioning. 
		whole = Ptr( new ItemGEMM( parts[ 0 ]->serialNumber, ratioWhole ) ) ;
		
		// all matrices wrap the same pre-allocated CPU memory, so just expand them
		expand( parts[ 0 ]->A.readCPU(),      whole->A.writeCPU()      ) ;
		expand( parts[ 0 ]->B.readCPU(),      whole->B.writeCPU()      ) ;
		expand( parts[ 0 ]->C.readCPU(),      whole->C.writeCPU()      ) ;
		expand( parts[ 0 ]->AB.readCPU(),     whole->AB.writeCPU()     ) ;
		expand( parts[ 0 ]->result.readCPU(), whole->result.writeCPU() ) ;
	}
	
} ; // class "ItemGEMM"

// source for fully initialized items
class SourceGEMM : public hawaii::flow::StepTemplate< hawaii::flow::ItemBase,
                                                      ItemGEMM               > {
	// c'tor with size of (square) matrices
	public:
	const int size ;
	SourceGEMM( const int sizeArg = 1024 ) :
		StepTemplate< hawaii::flow::ItemBase,
		              ItemGEMM               >( "GEMM source",          // name for printed evaluation
		                                        1,                      // use 1 CPU core
		                                        std::vector< int >(),   // don't use any GPUs
		                                        true,                   // source step
		                                        false                ), // no sink step
		size( sizeArg ) {
	}
	
	// inherited purely-virtual method
	public:
	void processImpl( const SrcPtr& srcPtr,
	                        DstPtr& dstPtr ) override {
		
		// copy serial number from source item
		dstPtr = DstPtr( new Dst( srcPtr->serialNumber ) ) ;
		
		// allocate all matrices
		dstPtr->A.writeCPU(      size, size, CV_32FC1 ) ;
		dstPtr->B.writeCPU(      size, size, CV_32FC1 ) ;
		dstPtr->C.writeCPU(      size, size, CV_32FC1 ) ;
		dstPtr->AB.writeCPU(     size, size, CV_32FC1 ) ;
		dstPtr->result.writeCPU( size, size, CV_32FC1 ) ;
		
		// set fixed dummy values to easily verify results: This can actually limit high-end systems.
		#if 0
			(cv::Mat&)dstPtr->A = cv::Scalar(   1.0f ) ;
			(cv::Mat&)dstPtr->B = cv::Scalar(   2.0f ) ;
			(cv::Mat&)dstPtr->C = cv::Scalar( 128.0f ) ;
		#endif
	}
} ;

// whole GEMM on CPU
class GEMMCPU : public hawaii::flow::StepTemplate< ItemGEMM > {
	
	// options as members, set name and cores used
	public:
	float alpha, beta ;
	GEMMCPU() :
		alpha( 0.25f ),
		beta(  0.75f ) {
		this->name = "CPU GEMM" ;
		this->setCPUCores( omp_get_max_threads() ) ;
	}
	
	// inherited purely-virtual method
	public:
	void processImpl( const SrcPtr& srcPtr,
	                        DstPtr& dstPtr ) override {
		
		// pass work item through
		dstPtr = srcPtr ;
		
		// wrap CPU function from "GEMM.h"
		gemmCPU( srcPtr->A, srcPtr->B, srcPtr->C,
		         dstPtr->result,
		         this->alpha, this->beta         ) ;
	}
} ;

// whole GEMM on GPU
class GEMMGPU : public hawaii::flow::StepTemplateGPU< ItemGEMM > {
	
	// options as members, CUBLAS support
	protected:
	cublasHandle_t handle ;
	public:
	float alpha, beta ;
	GEMMGPU() :
		alpha( 0.25f ),
		beta(  0.75f ) {
		this->name = "GPU GEMM" ;
		HAWAII_ERROR_NO_CUDA ;
		#ifndef HAWAII_NO_CUDA
			cublasCreate_v2( &handle ) ;
		#endif
	}
	~GEMMGPU() {
		HAWAII_ERROR_NO_CUDA ;
		#ifndef HAWAII_NO_CUDA
			cublasDestroy_v2( handle ) ;
		#endif
	}
	
	// inherited purely-virtual methods
	public:
	void uploadImpl( const SrcPtr&          srcPtr,
	                       SrcGPUPtr&       srcGPUPtr,
	                       cv::gpu::Stream& stream    ) override {
		
		// pass work item through
		srcGPUPtr = srcPtr ;
		
		// upload input to GPU
		srcGPUPtr->A.readGPU( stream ) ;
		srcGPUPtr->B.readGPU( stream ) ;
		srcGPUPtr->C.readGPU( stream ) ;
	}
	void computeImpl( const SrcGPUPtr&       srcGPUPtr,
	                        DstGPUPtr&       dstGPUPtr,
	                        cv::gpu::Stream& stream    ) override {
		
		// pass work item through
		dstGPUPtr = srcGPUPtr ;
		
		// wrap CPU function from "GEMM.h"
		gemmGPU( srcGPUPtr->A, srcGPUPtr->B, srcGPUPtr->C,
		         dstGPUPtr->result,
		         this->alpha, this->beta,
		         this->handle, stream                     ) ;
	}
	void downloadImpl( const DstGPUPtr&       dstGPUPtr,
	                         DstPtr&          dstPtr,
	                         cv::gpu::Stream& stream    ) override {
		
		// pass work item through
		dstPtr = dstGPUPtr ;
		
		// download result from GPU
		dstPtr->result.readCPU( stream ) ;
	}
} ;

// matrix multiplication on CPU
class MultiplyCPU : public hawaii::flow::StepTemplate< ItemGEMM > {
	
	// set name and cores used
	public:
	MultiplyCPU() {
		this->name = "CPU multiply" ;
		this->setCPUCores( omp_get_max_threads() ) ;
	}
	
	// inherited purely-virtual method
	public:
	void processImpl( const SrcPtr& srcPtr,
	                        DstPtr& dstPtr ) override {
		
		// pass work item through
		dstPtr = srcPtr ;
		
		// wrap CPU function from "GEMM.h"
		multiplyCPU( srcPtr->A, srcPtr->B,
		             dstPtr->AB           ) ;
	}
} ;

// matrix multiplication on GPU
class MultiplyGPU : public hawaii::flow::StepTemplateGPU< ItemGEMM > {
	
	// set name, CUBLAS support
	protected:
	cublasHandle_t handle ;
	public:
	MultiplyGPU() {
		this->name = "GPU multiply" ;
		HAWAII_ERROR_NO_CUDA ;
		#ifndef HAWAII_NO_CUDA
			cublasCreate_v2( &handle ) ;
		#endif
	}
	~MultiplyGPU() {
		HAWAII_ERROR_NO_CUDA ;
		#ifndef HAWAII_NO_CUDA
			cublasDestroy_v2( handle ) ;
		#endif
	}
	
	// inherited purely-virtual methods
	public:
	void uploadImpl( const SrcPtr&          srcPtr,
	                       SrcGPUPtr&       srcGPUPtr,
	                       cv::gpu::Stream& stream    ) override {
		
		// pass work item through
		srcGPUPtr = srcPtr ;
		
		// upload input to GPU
		srcGPUPtr->A.readGPU( stream ) ;
		srcGPUPtr->B.readGPU( stream ) ;
	}
	void computeImpl( const SrcGPUPtr&       srcGPUPtr,
	                        DstGPUPtr&       dstGPUPtr,
	                        cv::gpu::Stream& stream    ) override {
		
		// pass work item through
		dstGPUPtr = srcGPUPtr ;
		
		// wrap CPU function from "GEMM.h"
		multiplyGPU( srcGPUPtr->A, srcGPUPtr->B,
		             dstGPUPtr->AB,
		             this->handle, stream       ) ;
	}
	void downloadImpl( const DstGPUPtr&       dstGPUPtr,
	                         DstPtr&          dstPtr,
	                         cv::gpu::Stream& stream    ) override {
		
		// pass work item through
		dstPtr = dstGPUPtr ;
		
		// download result from GPU
		dstPtr->AB.readCPU( stream ) ;
	}
} ;

// matrix addition on CPU
class AddScaledCPU : public hawaii::flow::StepTemplate< ItemGEMM > {
	
	// options as members, set name and cores used
	public:
	float alpha, beta ;
	AddScaledCPU() :
		alpha( 0.25f ),
		beta(  0.75f ) {
		this->name = "CPU scaled add" ;
		this->setCPUCores( omp_get_max_threads() ) ;
	}
	
	// inherited purely-virtual method
	public:
	void processImpl( const SrcPtr& srcPtr,
	                        DstPtr& dstPtr ) override {
		
		// pass work item through
		dstPtr = srcPtr ;
		
		// wrap CPU function from "GEMM.h"
		addScaledCPU( srcPtr->AB, srcPtr->C,
		              dstPtr->result,
		              this->alpha, this->beta ) ;
	}
} ;

// matrix addition on GPU
class AddScaledGPU : public hawaii::flow::StepTemplateGPU< ItemGEMM > {
	
	// options as members
	public:
	float alpha, beta ;
	AddScaledGPU() :
		alpha( 0.25f ),
		beta(  0.75f ) {
		this->name = "GPU scaled add" ;
	}
	
	// inherited purely-virtual methods
	public:
	void uploadImpl( const SrcPtr&          srcPtr,
	                       SrcGPUPtr&       srcGPUPtr,
	                       cv::gpu::Stream& stream    ) override {
		
		// pass work item through
		srcGPUPtr = srcPtr ;
		
		// upload input to GPU
		srcGPUPtr->AB.readGPU( stream ) ;
	}
	void computeImpl( const SrcGPUPtr&       srcGPUPtr,
	                        DstGPUPtr&       dstGPUPtr,
	                        cv::gpu::Stream& stream    ) override {
		
		// pass work item through
		dstGPUPtr = srcGPUPtr ;
		
		// wrap CPU function from "GEMM.h"
		addScaledGPU( srcGPUPtr->AB, srcGPUPtr->C,
		              dstGPUPtr->result,
		              this->alpha, this->beta,
		              stream                      ) ;
	}
	void downloadImpl( const DstGPUPtr&       dstGPUPtr,
	                         DstPtr&          dstPtr,
	                         cv::gpu::Stream& stream    ) override {
		
		// pass work item through
		dstPtr = dstGPUPtr ;
		
		// download result from GPU
		dstPtr->result.readCPU( stream ) ;
	}
} ;

// sink to print a result's first entry
class SinkGEMM : public hawaii::flow::StepTemplate< ItemGEMM,
                                                    hawaii::flow::ItemBase > {
	// default c'tor
	public:
	SinkGEMM() :
		StepTemplate< ItemGEMM,
		              hawaii::flow::ItemBase >( "GEMM sink",             // name for printed evaluation
		                                        1,                       // use 1 CPU core
		                                        std::vector< int >(),    // don't use any GPUs
		                                        false,                   // no source step
		                                        true                 ) { // sink step
	}
	
	// inherited purely-virtual method
	public:
	void processImpl( const SrcPtr& srcPtr,
	                        DstPtr& dstPtr ) override {
		// print message
		hawaii::cout << "finished GEMM item #" << srcPtr->serialNumber << ", result( 0, 0 ) = " << srcPtr->result.readCPU().at< float >( 0, 0 ) << hawaii::endl ;
		
		// generate an output item for performance measurement even though this is a sink
		dstPtr = DstPtr( new Dst( srcPtr->serialNumber ) ) ;
	}
} ;
