// Copyright (C) 2013 by FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of demoHeterogeneousParallelism.
// 
// demoHeterogeneousParallelism is free software: you can redistribute it and/or modify it under the terms of the GNU 
// General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your 
// option) any later version.
// 
// demoHeterogeneousParallelism is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
// even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
// 
// You should have received a copy of the GNU General Public License along with demoHeterogeneousParallelism. If not, 
// see <http://www.gnu.org/licenses/>.


// simple demo on cooperatively using CPU and GPU with "libHawaii" classes
// =======================================================================

// work items and processing steps to wrap sample application code
#include "GEMMItemSteps.h"
#include "featureMatchingItemSteps.h"
#include "DoGSteps.h"
#include "dummyItemSteps.h"

// generic processing steps to load and display images
#include "hawaii/IO/load.h"
#include "hawaii/IO/display.h"

// "libHawaii" classes to combine processing steps for heterogeneous parallel computing
#include "hawaii/flow/series.h"
#include "hawaii/flow/pipeline.h"
#include "hawaii/flow/demandAlloc.h"
#include "hawaii/flow/partitioning.h"

// remaining headers
#include "hawaii/GPU/init.h"
#include <iostream>

// helpers only used here
namespace {
	
	// run a "hawaii::Flow" for a certain fixed time and regularly print its evaluation
	void runEvalStopFlow( hawaii::flow::FlowBase& flow,
	                      int                     seconds = 30,
	                      int                     step    = 10 ) {
		flow.run() ;
		for( ; seconds > 0 ; seconds -= step ) {
			sleep( step ) ;
			flow.eval() ;
		}
		flow.stop() ;
	}
}

// no command line arguments at this time
int main() {
	
	// show algorithmic results (e.g. drawn feature matches): This may limit performance, so do not use for benchmarking.
	const bool showResults = false ;
	
	// disable overlapping GPU computations and memory transfers: For shared-memory GPUs it is always disabled. This 
	//                                                            turns it off for discrete-memory GPUs as well.
	const bool disableOverlap = false ;
	
	// enable peer-to-peer GPU memory access, disable CPU busy-waiting for all GPUs, local scope for valid shutdown
	hawaii::GPU::init() ; {
	
	// choose application: 0 for general matrix multiplication
	//                     1 for feature detection/extraction and matching
	//                     2 for difference with Gaussian blur
	//                     3 for dummy sample application
	#define APPLICATION 0
	#if( APPLICATION == 0 )
		
		// source and sink processing steps
		SourceGEMM source( 1024 ) ;
		SinkGEMM   sink           ;
		
		// one processing step for all CPU cores
		MultiplyCPU  step1CPU     ;
		AddScaledCPU step2CPU     ;
		GEMMCPU      step1and2CPU ;
		
		// individual processing steps for each GPU
		MultiplyGPU  steps1GPU[     hawaii::GPUs ] ;
		AddScaledGPU steps2GPU[     hawaii::GPUs ] ;
		GEMMGPU      steps1and2GPU[ hawaii::GPUs ] ;
		for( int GPU = 0 ; GPU < hawaii::GPUs ; ++GPU ) {
			std::stringstream suffix ; suffix << " (on GPU #" << GPU << ")" ;
			steps1GPU[     GPU ].name += suffix.str() ;
			steps2GPU[     GPU ].name += suffix.str() ;
			steps1and2GPU[ GPU ].name += suffix.str() ;
			steps1GPU[     GPU ].setGPU( GPU ) ;
			steps2GPU[     GPU ].setGPU( GPU ) ;
			steps1and2GPU[ GPU ].setGPU( GPU ) ;
			steps1GPU[     GPU ].overlapKernelsAndCopies = !disableOverlap ;
			steps2GPU[     GPU ].overlapKernelsAndCopies = !disableOverlap ;
			steps1and2GPU[ GPU ].overlapKernelsAndCopies = !disableOverlap ;
		}
		
		// (dummy) partitioning information
		hawaii::flow::PartInfoBase partInfo ;
		
	#elif( APPLICATION == 1 )
		
		// source and sink processing steps, intermediate buffer for previous keypoints and descriptors
		hawaii::IO::StepLoad  source( "data/%03llu.png", "orig", true ) ;
		BufferFeatureMatching buffer                                    ;
		SinkFeatureMatching   sink                                      ;
		
		// one processing step for all CPU cores: combine the buffer with the first step
		DetectExtractCPU     step1UnbufferedCPU                                   ;
		hawaii::flow::Series step1CPU(     step1UnbufferedCPU, buffer           ) ;
		MatchCPU             step2CPU                                             ;
		hawaii::flow::Series step1and2CPU( step1UnbufferedCPU, buffer, step2CPU ) ;
		step1CPU.name     = "CPU detect/extract and buffer"        ;
		step1and2CPU.name = "CPU detect/extract, buffer and match" ;
		
		// individual processing steps for each GPU
		DetectExtractGPU     steps1UnbufferedGPU[ hawaii::GPUs ] ;
		hawaii::flow::Series steps1GPU[           hawaii::GPUs ] ;
		MatchGPU             steps2GPU[           hawaii::GPUs ] ;
		hawaii::flow::Series steps1and2GPU[       hawaii::GPUs ] ;
		for( int GPU = 0 ; GPU < hawaii::GPUs ; ++GPU ) {
			std::stringstream suffix ; suffix << " (on GPU #" << GPU << ")" ;
			steps1GPU[           GPU ].name = "GPU detect/extract and buffer" ;
			steps1and2GPU[       GPU ].name = "GPU detect/extract, buffer and match" ;
			steps1UnbufferedGPU[ GPU ].name += suffix.str() ;
			steps1GPU[           GPU ].name += suffix.str() ;
			steps2GPU[           GPU ].name += suffix.str() ;
			steps1and2GPU[       GPU ].name += suffix.str() ;
			steps1UnbufferedGPU[ GPU ].setGPU( GPU ) ;
			steps2GPU[           GPU ].setGPU( GPU ) ;
			steps2GPU[           GPU ].overlapKernelsAndCopies = !disableOverlap ;
			steps1GPU[           GPU ].add( steps1UnbufferedGPU[ GPU ], buffer                   ) ;
			steps1and2GPU[       GPU ].add( steps1UnbufferedGPU[ GPU ], buffer, steps2GPU[ GPU ] ) ;
		}
		
		// partitioning information
		hawaii::flow::PartInfoImageOverlap partInfo( 1 + 2 * step1UnbufferedCPU.options.edgeThreshold ) ;
		
	#elif( APPLICATION == 2 )
		
		// source and sink processing steps
		hawaii::IO::StepLoad    source( "data/%03llu.png", "orig", true ) ;
		hawaii::IO::StepDisplay sink                                      ;
		
		// one processing step for all CPU cores
		GaussCPU             step1CPU                           ;
		DiffCPU              step2CPU                           ;
		hawaii::flow::Series step1and2CPU( step1CPU, step2CPU ) ;
		step1and2CPU.name = "CPU blur and diff" ;
		
		// individual processing steps for each GPU
		GaussGPU     steps1GPU[     hawaii::GPUs ] ;
		DiffGPU      steps2GPU[     hawaii::GPUs ] ;
		GaussDiffGPU steps1and2GPU[ hawaii::GPUs ] ;
		for( int GPU = 0 ; GPU < hawaii::GPUs ; ++GPU ) {
			std::stringstream suffix ; suffix << " (on GPU #" << GPU << ")" ;
			steps1GPU[     GPU ].name += suffix.str() ;
			steps2GPU[     GPU ].name += suffix.str() ;
			steps1and2GPU[ GPU ].name += suffix.str() ;
			steps1GPU[     GPU ].setGPU( GPU ) ;
			steps2GPU[     GPU ].setGPU( GPU ) ;
			steps1and2GPU[ GPU ].setGPU( GPU ) ;
			steps1GPU[     GPU ].overlapKernelsAndCopies = !disableOverlap ;
			steps2GPU[     GPU ].overlapKernelsAndCopies = !disableOverlap ;
			steps1and2GPU[ GPU ].overlapKernelsAndCopies = !disableOverlap ;
		}
		
		// partitioning information
		hawaii::flow::PartInfoImageOverlap partInfo ;
		partInfo.addPreAlloc( "blur", "orig", hawaii::flow::PartInfoImageOverlap::typeSame, step1CPU.win ) ;
		partInfo.addPreAlloc( "diff", "blur"                                                             ) ;
		
	#elif( APPLICATION == 3 )
		
		// source and sink processing steps
		SourceDummy source ;
		SinkDummy   sink   ;
		
		// one processing step for all CPU cores
		FirstStepDummyCPU  step1CPU ;
		SecondStepDummyCPU step2CPU ;
		hawaii::flow::Series step1and2CPU( step1CPU, step2CPU ) ;
		step1and2CPU.name = "CPU both dummy steps" ;
		
		// individual processing steps for each GPU
		FirstStepDummyGPU    steps1GPU[     hawaii::GPUs ] ;
		SecondStepDummyGPU   steps2GPU[     hawaii::GPUs ] ;
		hawaii::flow::Series steps1and2GPU[ hawaii::GPUs ] ;
		for( int GPU = 0 ; GPU < hawaii::GPUs ; ++GPU ) {
			std::stringstream suffix ; suffix << " (on GPU #" << GPU << ")" ;
			steps1and2GPU[ GPU ].name = "GPU both dummy steps" ;
			steps1GPU[     GPU ].name += suffix.str() ;
			steps2GPU[     GPU ].name += suffix.str() ;
			steps1and2GPU[ GPU ].name += suffix.str() ;
			steps1GPU[     GPU ].setGPU( GPU ) ;
			steps2GPU[     GPU ].setGPU( GPU ) ;
			steps1GPU[     GPU ].overlapKernelsAndCopies = !disableOverlap ;
			steps2GPU[     GPU ].overlapKernelsAndCopies = !disableOverlap ;
			steps1and2GPU[ GPU ].add( steps1GPU[ GPU ], steps2GPU[ GPU ] ) ;
		}
		
		// (no) partitioning information
		hawaii::flow::PartInfoBase partInfo ;
		
	#endif
	
	// CPU series: As a baseline for benchmarking, processing steps are called on CPU one after another.
	if( 1 ) {
		
		// combine steps and queues into pipeline, optionally add displaying step
		hawaii::flow::Pipeline pipe( "pipeline around CPU series",
		                             source,
		                             hawaii::flow::queueFactory( hawaii::flow::queueTypeFIFO ),
		                             step1and2CPU                                              ) ;
		if( showResults ) {
			pipe.add( hawaii::flow::queueFactory( hawaii::flow::queueTypeFIFOOverflow ),
			          sink                                                              ) ;
		}
		
		// set throughput limit to measure reduced power consumption
		if( 0 ) { source.setThroughputLimit( 10.0 ) ; }
		
		// run the flow for a fixed time, regularly print evaluation
		runEvalStopFlow( pipe ) ;
	}
	
	// GPU series: see above
	if( 1 && ( hawaii::GPUs > 0 ) ) {
		
		// combine steps and queues into pipeline, optionally add displaying step
		hawaii::flow::Pipeline pipe( "pipeline around GPU series",
		                             source,
		                             hawaii::flow::queueFactory( hawaii::flow::queueTypeFIFO ),
		                             steps1and2GPU[ 0 ]                                        ) ;
		if( showResults ) {
			pipe.add( hawaii::flow::queueFactory( hawaii::flow::queueTypeFIFOOverflow ),
			          sink                                                              ) ;
		}
		
		// set throughput limit to measure reduced power consumption
		if( 0 ) { source.setThroughputLimit( 10.0 ) ; }
		
		// run the flow for a fixed time, regularly print evaluation
		runEvalStopFlow( pipe ) ;
	}
	
	// pipeline: The first processing step continuously runs on either CPU or GPU, while at the same time the second 
	//           step uses the other processor.
	if( 1 && ( hawaii::GPUs > 0 ) ) {
		
		// combine steps within inner pipeline to get its performance statistics
		hawaii::flow::Pipeline pipeInner( "heterogeneous pipeline",
		                                  #if APPLICATION != 1
		                                  steps1GPU[ 0 ],
		                                  hawaii::flow::queueFactory( hawaii::flow::queueTypeFIFO ),
		                                  step2CPU                                                  ) ;
		                                  #else
		                                  step1CPU,
		                                  hawaii::flow::queueFactory( hawaii::flow::queueTypeFIFO ),
		                                  steps2GPU[ 0 ]                                            ) ;
		                                  #endif
		
		// add source and optional sink via outer pipeline
		hawaii::flow::Pipeline pipeOuter( "pipeline around heterogeneous pipeline",
		                                  source,
		                                  hawaii::flow::queueFactory( hawaii::flow::queueTypeFIFO ),
		                                  pipeInner                                                 ) ;
		if( showResults ) {
			pipeOuter.add( hawaii::flow::queueFactory( hawaii::flow::queueTypeFIFOOverflow ),
			               sink                                                              ) ;
		}
		
		// set throughput limit to measure reduced power consumption
		if( 0 ) { source.setThroughputLimit( 10.0 ) ; }
		
		// run the flow for a fixed time, regularly print evaluation
		runEvalStopFlow( pipeOuter ) ;
	}
	
	// demand-based allocation: Compute both processing steps on the first processor to become available after a work 
	//                          item has been loaded.
	if( 1 ) {
		
		// combine processing steps for all processors
		#if APPLICATION != 1
			hawaii::flow::DemandAlloc demal( "demand-based allocation", step1and2CPU ) ;
			demal.setFlowPreceding( source ) ;
			for( int GPU = 0 ; GPU < hawaii::GPUs ; ++GPU ) {
				demal.add( steps1and2GPU[ GPU ] ) ;
			}
		#else
			hawaii::flow::DemandAlloc demal1( "demand-based allocation of step 1", step1CPU ) ;
			hawaii::flow::DemandAlloc demal2( "demand-based allocation of step 2", step2CPU ) ;
			demal1.setFlowPreceding( source ) ;
			demal2.setFlowPreceding( buffer ) ;
			if( hawaii::GPUs > 0 ) {
				demal1.add( steps1UnbufferedGPU[ 0 ] ) ;
			}
			for( int GPU = 0 ; GPU < hawaii::GPUs ; ++GPU ) {
				demal2.add( steps2GPU[ GPU ] ) ;
			}
		#endif
		
		// add source and optional sink via pipeline
		hawaii::flow::Pipeline pipe( "pipeline around demand-based allocation",
		                             source,
		                             hawaii::flow::queueFactory( hawaii::flow::queueTypeFIFO    ),
		                             #if APPLICATION != 1
		                             demal                                                        ) ;
		                             #else
		                             demal1,
		                             hawaii::flow::queueFactory( hawaii::flow::queueTypeOrdered,
		                                                         1 + hawaii::GPUs               ),
		                             buffer,
		                             hawaii::flow::queueFactory( hawaii::flow::queueTypeFIFO    ),
		                             demal2                                                       ) ;
		                             #endif
		if( showResults ) {
			pipe.add( hawaii::flow::queueFactory( hawaii::flow::queueTypeOrderedOverflow ),
			          sink                                                                 ) ;
		}
		
		// set throughput limit to optimize power efficiency instead of performance
		if( 0 ) { source.setThroughputLimit( 10.0 ) ; }
		
		// run the flow for a fixed time, regularly print evaluation
		runEvalStopFlow( pipe ) ;
	}
	
	// partitioning: Adaptively split input work item into partitions, each of which will be processed concurrently on 
	//               an individual processor.
	if( 1 ) {
		
		// combine processing steps for all processors
		#if APPLICATION != 1
			hawaii::flow::Partitioning part( "partitioning", partInfo, step1and2CPU ) ;
			part.setFlowPreceding( source ) ;
			for( int GPU = 0 ; GPU < hawaii::GPUs ; ++GPU ) {
				part.add( steps1and2GPU[ GPU ] ) ;
			}
		#else
			hawaii::flow::Partitioning part1( "partitioning of step 1", partInfo, step1UnbufferedCPU ) ;
			hawaii::flow::Partitioning part2( "partitioning of step 2", partInfo, step2CPU           ) ;
			part1.setFlowPreceding( source ) ;
			part2.setFlowPreceding( buffer ) ;
			part1.ratioMin = 0.25 ; // developer note: An assertion in "cv::gpu::ORB_GPU" fails if the partition is too small.
			if( hawaii::GPUs > 0 ) {
				part1.add( steps1UnbufferedGPU[ 0 ] ) ;
			}
			for( int GPU = 0 ; GPU < hawaii::GPUs ; ++GPU ) {
				part2.add( steps2GPU[ GPU ] ) ;
			}
		#endif
		
		// add source and sink display via pipeline
		hawaii::flow::Pipeline pipe( "pipeline around partitioning",
		                             source,
		                             hawaii::flow::queueFactory( hawaii::flow::queueTypeFIFO ),
		                             #if APPLICATION != 1
		                             part                                                      ) ;
		                             #else
		                             part1,
		                             hawaii::flow::queueFactory( hawaii::flow::queueTypeFIFO ),
		                             buffer,
		                             hawaii::flow::queueFactory( hawaii::flow::queueTypeFIFO ),
		                             part2                                                     ) ;
		                             #endif
		if( showResults ) {
			pipe.add( hawaii::flow::queueFactory( hawaii::flow::queueTypeFIFOOverflow ),
			          sink                                                              ) ;
		}
		
		// set throughput limit to optimize power efficiency instead of performance
		if( 0 ) { source.setThroughputLimit( 10.0 ) ; }
		
		// run the flow for a fixed time, regularly print evaluation
		runEvalStopFlow( pipe ) ;
	}
	
	// local scope for valid shutdown, clear memory pools, flush CUDA profiler information
	} hawaii::GPU::reset() ;
	return 0 ;
}
