// Copyright (C) 2013 by FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of libHawaii.
// 
// libHawaii is free software: you can redistribute it and/or modify it under the terms of the GNU General Public 
// License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later 
// version.
// 
// libHawaii is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with libHawaii. If not, see 
// <http://www.gnu.org/licenses/>.


// pools to avoid the latency and synchronization caused by "cudaMalloc*()" and "cudaHostAlloc()" 
// ==============================================================================================

#pragma once

#include "hawaii/GPU/openCV.h"
#include "hawaii/common/hardware.h"
#include "hawaii/common/sync.h"
#include "hawaii/GPU/nvccFixHeaderAbove.h"
#include <boost/thread/mutex.hpp>
#include "hawaii/GPU/nvccFixHeaderBelow.h"
#include <string>

namespace hawaii {
namespace GPU {

// common pool for "cv::gpu::GpuMat" (both in actual GPU and in mapped CPU memory) and "cv::gpu::CudaMem"
// user note: just use the global instance "hawaii::GPU::memoryPool"
class MemoryPool {
	
	// replacement for "cv::gpu::GpuMat::create()": additionally allows specifying memory type (see above), modifies 
	//                                              the given instance only if necessary
	public:
	void operator ()( cv::gpu::GpuMat& gpuMatArg,
	                  const int rows, const int cols, const int type,
	                  MemoryType memory = memoryAuto ) ;
	void operator ()( cv::gpu::GpuMat& gpuMatArg,
	                  const cv::Size size,            const int type,
	                  MemoryType memory = memoryAuto ) {
		this->operator ()( gpuMatArg,
		                   size.height, size.width, type,
		                   memory ) ;
	}
	
	// replacement for "cv::gpu::CudaMem::create()": makes the given instance page-locked and mappable into GPU 
	//                                               address space (but not write-combined), modifies it only if 
	//                                               necessary
	public:
	void operator ()( cv::gpu::CudaMem& cudaMemArg,
	                  const int rows, const int cols, const int type ) ;
	void operator ()( cv::gpu::CudaMem& cudaMemArg,
	                  const cv::Size size,            const int type ) {
		this->operator ()( cudaMemArg,
		                   size.height, size.width, type ) ;
	}
	
	// d'tor
	// user note: Global pools must be explicitly cleared because CUDA may already be unloaded before their d'tors are 
	//            executed.
	// developer note: The default d'tor is insufficient because pooled instances must be destroyed while their 
	//                 corresponding GPU is active.
	public:
	void clear() ;
	~MemoryPool() ;
	
	// print number of available and lent items of each (non-empty) internal pool
	public:
	std::string statusString() const ;
	
	// actual pools: Each of the 64 size bins contains allocations of "2^n" bytes and serves requests between 
	//               "2^(n-1)+1" and "2^n" bytes. Additionally, pools for "cv::gpu::GpuMat" are device-specific 
	//               while those for "cv::gpu::CudaMem" are not.
	// developer note #1: When requesting a "cv::gpu::GpuMat", own GPU memory will come from a "PoolGpuMat" but 
	//                    mapped CPU memory will come from a "PoolCudaMem".
	// developer note #2: "std::pair::first" lock the pool's mutex, "std::pair::second" use the pool.
	protected:
	typedef hawaii::SyncPair< boost::mutex, std::vector< cv::gpu::GpuMat  > > PoolGpuMat  ;
	typedef hawaii::SyncPair< boost::mutex, std::vector< cv::gpu::CudaMem > > PoolCudaMem ;
	PoolGpuMat  poolsGpuMat[  HAWAII_GPUS_MAX ][ 64 ] ;
	PoolCudaMem poolsCudaMem[                    64 ] ;
	
} ; // class "MemoryPool"

// global instance
// user note: If you have used this instance, call "hawaii::GPU::reset()" or "hawaii::GPU::memoryPool.clear()" at the 
//            end of your program (see note at "MemoryPool::clear()" for details).
extern MemoryPool memoryPool ;

} } // namespaces "hawaii::GPU"
