// Copyright (C) 2013 by FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of libHawaii.
// 
// libHawaii is free software: you can redistribute it and/or modify it under the terms of the GNU General Public 
// License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later 
// version.
// 
// libHawaii is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with libHawaii. If not, see 
// <http://www.gnu.org/licenses/>.


// processing step to load images from disk
// ========================================

#pragma once

#include "hawaii/GPU/nvccProhibit.h"
#include "hawaii/flow/step.h"
#include "hawaii/flow/itemNamedImages.h"

namespace hawaii {
namespace IO {

// load image with filename generated from pattern and index
cv::Mat loadIndexed( const std::string pattern,
                     const uint64_t    index   ) ;

// processing step to load images from disk
class StepLoad : public hawaii::flow::StepTemplate< hawaii::flow::ItemBase,
                                                    hawaii::flow::ItemNamedImages > {
	
	// c'tor to optionally set any option for which "add()" below is defined
	public:
	template< typename ... Args >
	StepLoad( Args&& ... args ) :
		hawaii::flow::StepTemplate< hawaii::flow::ItemBase,
		                            hawaii::flow::ItemNamedImages >( "load images from disk",
		                                                             1, std::vector< int >(),
		                                                             true, false ),
		indexCurr(  0                                      ),
		indexStart( 0                                      ),
		indexEnd(   std::numeric_limits< uint64_t >::max() ),
		bufferAndRepeat( false ) {
		this->add( std::forward< Args >( args ) ... ) ;
	}
	
	// reset internal state
	public:
	virtual void reset() override {
		this->indexCurr = this->indexStart ;
		this->bufferedItems.clear() ;
		StepBase::reset() ;
	}
	
	// add one pair of filename pattern and instance name at a time
	// user note #1: The filename pattern must contain the filename index as an unsigned 64-bit integer variable with or 
	//               without leading zeros and a user-defined number of digits, e.g. "myImages/input%010llu.png".
	// user note #2: The instance name is used to assign each loaded image to "dstPtr->images[ instanceName ]".
	// developer note: specific "(const )char*" versions needed because they would have been converted to "bool" otherwise
	protected:
	std::vector< std::pair< std::string, std::string > > imageDescs ;
	public:
	template< typename ... Args >
	void add( const std::string& filePatternNext,
	          const std::string& instanceNameNext,
	                Args&& ...   argsRest         ) {
		this->imageDescs.push_back( { filePatternNext, instanceNameNext } ) ;
		this->add( std::forward< Args >( argsRest ) ... ) ;
	}
	template< typename ... Args >
	void add( const char*      filePatternNext,
	          const char*      instanceNameNext,
	                Args&& ... argsRest         ) {
		this->imageDescs.push_back( { std::string( filePatternNext ), std::string( instanceNameNext ) } ) ;
		this->add( std::forward< Args >( argsRest ) ... ) ;
	}
	template< typename ... Args >
	void add( char*      filePatternNext,
	          char*      instanceNameNext,
	          Args&& ... argsRest         ) {
		this->imageDescs.push_back( { std::string( filePatternNext ), std::string( instanceNameNext ) } ) ;
		this->add( std::forward< Args >( argsRest ) ... ) ;
	}
	
	// set start (inclusive) and end (exclusive) of filename index range
	// developer note: specific "int" version needed because they would have been converted to "bool" otherwise
	protected:
	uint64_t indexCurr  ;
	uint64_t indexStart ;
	uint64_t indexEnd   ;
	template< typename ... Args >
	void add( const uint64_t   indexStartArg,
	          const uint64_t   indexEndArg,
	                Args&& ... argsRest      ) {
		this->indexCurr  = indexStartArg ;
		this->indexStart = indexStartArg ;
		this->indexEnd   = indexEndArg   ;
		this->add( std::forward< Args >( argsRest ) ... ) ;
	}
	template< typename ... Args >
	void add( const int        indexStartArg,
	          const int        indexEndArg,
	                Args&& ... argsRest      ) {
		this->indexCurr  = indexStartArg ;
		this->indexStart = indexStartArg ;
		this->indexEnd   = indexEndArg   ;
		this->add( std::forward< Args >( argsRest ) ... ) ;
	}
	
	// instead of loading and passing one set of images at a time and finally passing a special "final" work item, pre- 
	// load all image sets to a buffer and then repeatedly pass its content
	// user note: leave at default "false" for processing a data set and saving/evaluating results, set "true" for 
	//            benchmarking or a continuously running demo
	protected:
	bool bufferAndRepeat ;
	mutable hawaii::flow::ItemNamedImages::PtrVec bufferedItems ;
	template< typename ... Args >
	void add( const bool       bufferAndRepeatArg,
	                Args&& ... argsRest           ) {
		this->bufferAndRepeat = bufferAndRepeatArg ;
		this->add( std::forward< Args >( argsRest ) ... ) ;
	}
	public:
	
	// allow the variadic template "add()" to terminate
	protected:
	void add() const {}
	
	// implement inherited purely-virtual method to load the images
	public:
	virtual void processImpl( const SrcPtr& srcPtr,
	                                DstPtr& dstPtr ) override ;
} ; // class "StepLoad"

} } // namespaces "hawaii::IO"
