// Copyright (C) 2013 by FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of libHawaii.
// 
// libHawaii is free software: you can redistribute it and/or modify it under the terms of the GNU General Public 
// License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later 
// version.
// 
// libHawaii is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with libHawaii. If not, see 
// <http://www.gnu.org/licenses/>.


// unit test helpers
// =================

#pragma once

#include "hawaii/GPU/CUDA.h"
#include "hawaii/common/timer.h"
#include "hawaii/common/hardware.h"
#include "hawaii/common/ostream.h"
#include <opencv2/core/core.hpp>
#include <string>

// perform a call a given number of times, then print the average latency
#define HAWAII_BENCHMARK_CPU( CALL, RUNS ) { \
	printf( "benchmarking %s...\n", #CALL ) ; \
	hawaii::Timer timer( true ) ; \
	for( size_t run = 0 ; run < RUNS ; ++run ) { CALL ; } \
	printf( "%f ms/call\n", timer.toc() / RUNS ) ; \
}

// same as "BENCHMARK_CPU()", but try a "cores" parameter from 1 to the total number of threads (incl. "Hyper Threading)
#define HAWAII_BENCHMARK_CPU_CORES( CALL, RUNS ) { \
	printf( "benchmarking %s...\n", #CALL ) ; \
	for( int cores = 1 ; cores <= hawaii::CPUCores ; ++cores ) { \
		hawaii::Timer timer( true ) ; \
		for( size_t run = 0 ; run < RUNS ; ++run ) { CALL ; } \
		printf( "%02d cores: %f ms/call\n", cores, timer.toc() / RUNS ) ; \
	} \
}

// same as "BENCHMARK_CPU()", but try a "SIMD" parameter with both "true" and "false"  
#define HAWAII_BENCHMARK_CPU_SIMD( CALL, RUNS ) { \
	printf( "benchmarking %s...\n", #CALL ) ; \
		bool SIMD = false ; \
		hawaii::Timer timer( true ) ; \
		for( size_t run = 0 ; run < RUNS ; ++run ) { CALL ; } \
		const double timeScalar = timer.toc() / RUNS ; \
		SIMD = true ; \
		timer.tic() ; \
		for( size_t run = 0 ; run < RUNS ; ++run ) { CALL ; } \
		const double timeSIMD = timer.toc() / RUNS ; \
		printf( "%f (scalar) / %f (SIMD) ms/call (speed-up: %f)\n", timeScalar, timeSIMD, timeScalar/timeSIMD ) ; \
}

// combination of "BENCHMARK_CPU_CORES()" and "BENCHMARK_CPU_SIMD()"
#define HAWAII_BENCHMARK_CPU_CORES_SIMD( CALL, RUNS ) { \
	printf( "benchmarking %s...\n", #CALL ) ; \
	for( int cores = 1 ; cores <= hawaii::CPUCores ; ++cores ) { \
		bool SIMD = false ; \
		hawaii::Timer timer( true ) ; \
		for( size_t run = 0 ; run < RUNS ; ++run ) { CALL ; } \
		const double timeScalar = timer.toc() / RUNS ; \
		SIMD = true ; \
		timer.tic() ; \
		for( size_t run = 0 ; run < RUNS ; ++run ) { CALL ; } \
		const double timeSIMD = timer.toc() / RUNS ; \
		printf( "%02d cores: %f (scalar) / %f (SIMD) ms/call (speed-up: %f)\n", cores, timeScalar, timeSIMD, timeScalar/timeSIMD ) ; \
	} \
}

// same as "BENCHMARK_CPU()", but do one warm-up call to compile PTX code if necessary and synchronize the GPU before 
// stopping the timer (in case of an asynchronous call)
#define HAWAII_BENCHMARK_GPU( CALL, RUNS ) { \
	printf( "benchmarking %s...\n", #CALL ) ; \
	cv::gpu::Stream stream ; \
	CALL ; \
	stream.waitForCompletion() ; \
	cudaDeviceSynchronize() ; \
	hawaii::Timer timer( true ) ; \
	for( size_t run = 0 ; run < RUNS ; ++run ) { CALL ; } \
	stream.waitForCompletion() ; \
	cudaDeviceSynchronize() ; \
	printf( "%f ms/call\n", timer.toc() / RUNS ) ; \
}

// combination of "BENCHMARK_CPU_SIMD()" and "BENCHMARK_GPU()"
#define HAWAII_BENCHMARK_GPU_SIMD( CALL, RUNS ) { \
	printf( "benchmarking %s...\n", #CALL ) ; \
	cv::gpu::Stream stream ; \
	bool SIMD = false ; \
	CALL ; \
	stream.waitForCompletion() ; \
	cudaDeviceSynchronize() ; \
	hawaii::Timer timer( true ) ; \
	for( size_t run = 0 ; run < RUNS ; ++run ) { CALL ; } \
	stream.waitForCompletion() ; \
	cudaDeviceSynchronize() ; \
	const double timeScalar = timer.toc() / RUNS ; \
	SIMD = true ; \
	CALL ; \
	stream.waitForCompletion() ; \
	cudaDeviceSynchronize() ; \
	timer.tic() ; \
	for( size_t run = 0 ; run < RUNS ; ++run ) { CALL ; } \
	stream.waitForCompletion() ; \
	cudaDeviceSynchronize() ; \
	const double timeSIMD = timer.toc() / RUNS ; \
	printf( "%f (scalar) / %f (SIMD) ms/call (speed-up: %f)\n", timeScalar, timeSIMD, timeScalar/timeSIMD ) ; \
}

namespace hawaii {

// print whether a unit test's result is correct or not 
inline void printUnitTest( const std::string name,
                           const bool        correct ) {
	if( correct ) { hawaii::cout << name << " succeeded" << hawaii::endl ; }
	else          { hawaii::cerr << name << " FAILED"    << hawaii::endl ; }
}

// compare images pixel by pixel and print results
// developer note: We need the wrapper below because NVCC does not yet support C++11, particularly default template 
//                 arguments of function templates.
template< typename ElemA, typename ElemB >
inline bool printImageComparison( const cv::Mat imgA,
                                  const cv::Mat imgB ) {
	
	// compare dimensions
	if( imgA.rows       != imgB.rows       ) { hawaii::cerr << "unequal number of rows"     << hawaii::endl ; return false ; }
	if( imgA.cols       != imgB.cols       ) { hawaii::cerr << "unequal number of columns"  << hawaii::endl ; return false ; }
	if( imgA.channels() != imgB.channels() ) { hawaii::cerr << "unequal number of channels" << hawaii::endl ; return false ; }
	
	// statistics to be determined
	uint64_t count = 0, total = imgA.rows * imgA.cols * imgA.channels() ;
	double   error = 0.0 ;
	
	// check all pixels
	for( int row = 0 ; row < imgA.rows ; ++row ) {
		const ElemA* __restrict__ const rowA = imgA.ptr< ElemA >( row ) ;
		const ElemB* __restrict__ const rowB = imgB.ptr< ElemB >( row ) ;
		for( int colChn = 0 ; colChn < imgA.cols * imgA.channels() ; ++colChn ) {
			if( rowA[ colChn ] != rowB[ colChn ] ) {
				++count ;
				error += fabs( rowA[ colChn ] - rowB[ colChn ] ) ;
			}
		}
	}
	
	// print statistics
	if( count != 0 ) { hawaii::cerr << "different images: "
	                                << 100.0 * (double)count / (double)total << "% of pixels affected, "
	                                << error / (double)count                 << " mean per-pixel error"  << hawaii::endl ; return true  ; }
	else             { hawaii::cout << "identical images"                                                << hawaii::endl ; return false ; }
}
template< typename Elem >
inline bool printImageComparison( const cv::Mat imgA,
                                  const cv::Mat imgB ) {
	return printImageComparison< Elem, Elem >( imgA, imgB ) ;
}

} // namespace "hawaii"
