// Copyright (C) 2013 by FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of libHawaii.
// 
// libHawaii is free software: you can redistribute it and/or modify it under the terms of the GNU General Public 
// License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later 
// version.
// 
// libHawaii is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with libHawaii. If not, see 
// <http://www.gnu.org/licenses/>.


// demand-based allocation of work items to available processing units
// ===================================================================

#pragma once

#include "hawaii/GPU/nvccProhibit.h"
#include "hawaii/common/sync.h"
#include "hawaii/flow/flow.h"
#include <boost/thread/thread.hpp>
#include <boost/serialization/shared_ptr.hpp>
#include <set>
#include <atomic>

namespace hawaii {
namespace flow {

// wrapper around identical functionality on different devices
class DemandAlloc : public FlowBase {
	
	// c'tor to directly add any number of flow components and other arguments for which "DemandAlloc::add()" is defined
	public:
	template< typename ... Args >
	DemandAlloc( Args&& ... args ) ;
	
	// add one flow component at a time
	public:
	template< typename ... Args >
	void add( hawaii::flow::FlowBase& flowNext,
	          Args&& ...              argsRest ) ;
	
	// allow the variadic template to set the class's name and to terminate
	protected:
	template< typename ... Args >
	void add( const std::string nameArg,
	                Args&& ...  argsRest ) ;
	void add() {}
	
	// reset internal state
	public:
	virtual void reset() ;
	
	// contain pointers to flow components instead of copies, know preceding flow
	protected:
	FlowBase::PtrVec flows ;
	const FlowBase* flowPullPtr ;
	public:
	void setFlowPreceding( const FlowBase& flowPrecedingArg ) { this->flowPullPtr = &flowPrecedingArg ; }
	
	// run this demand-based allocator, wait for it to finish or directly stop it
	public:
	virtual void run()  override ;
	virtual void wait() override ;
	virtual void stop() override ;
	virtual ~DemandAlloc() { this->wait() ; }
	
	// find whether this demand-based allocator is a source and/or a sink
	protected:
	bool source, sink ;
	public:
	virtual bool isSource() const override { return this->source ; }
	virtual bool isSink()   const override { return this->sink   ; }
	
	// optionally limit throughput: You may get or set "INFINITY" for unlimited throughput.
	// developer note: This class enforces it internally rather than in "FlowBase" for being able to optimize power 
	//                 efficiency (see below). It is however less precise.
	protected:
	double throughputLimitDemandAlloc ;
	public:
	virtual double getThroughputLimit() const { return this->throughputLimitDemandAlloc ; }
	virtual void   setThroughputLimit( const double throughputLimit ) { this->throughputLimitDemandAlloc = throughputLimit ; }
	
	// time interval between throughput limit updates
	public:
	double timeUpdate ;
	protected:
	hawaii::Timer timerUpdate ;
	
	// estimate power consumption, update contained flows' throughput limits to optimize power efficiency
	public:
	virtual double power() const override ;
	protected:
	void updateThroughputLimits() ;
	
	// print evaluation
	public:
	virtual void eval( size_t indentationName = 0,
	                   size_t indentationData = 0 ) const override ;
	
	// override pulling interface to distribute final work item to all flow components
	protected:
	std::atomic_bool pullFinalItem ;
	virtual void pullImpl(    ItemBase::Ptr& itemPtr ) override ;
	virtual bool pullTryImpl( ItemBase::Ptr& itemPtr ) override ;
	
	// override pushing interface to retain final work item until all flow components have tried to push one
	protected:
	hawaii::SyncPair< boost::mutex, std::set< boost::thread::id > > pushedFinalItem ;
	virtual void pushImpl(    const ItemBase::Ptr& itemPtr ) override ;
	virtual bool pushTryImpl( const ItemBase::Ptr& itemPtr ) override ;
	
} ; // class "DemandAlloc"

// "DemandAlloc": wrapper around identical functionality on different devices
// --------------------------------------------------------------------------

// c'tor to directly add any number of flow components and other arguments for which "Series::add()" is defined
template< typename ... Args >
inline DemandAlloc::DemandAlloc( Args&& ... args ) :
	FlowBase( "demand-based allocator (unnamed)" ),
	flowPullPtr( nullptr ),
	source( false ),
	sink(   false ),
	throughputLimitDemandAlloc( INFINITY ),
	timeUpdate( 1000.0 ),
	timerUpdate( false ),
	pullFinalItem( false ) {
	this->add( std::forward< Args >( args ) ... ) ;
}

// add any number of flow components at once
template< typename ... Args >
inline void DemandAlloc::add( hawaii::flow::FlowBase& flowNext,
                              Args&& ...              argsRest ) {
	
	// copy source/sink characteristics from first component
	if( this->flows.empty() ) {
		this->source = flowNext.isSource() ;
		this->sink   = flowNext.isSink()   ; }
		
	// demand same characteristics from other components
	else {
		HAWAII_ERROR_CONDITIONAL( this->isSource() != flowNext.isSource()
		                       || this->isSink()   != flowNext.isSink(),
		                          "incompatible source/sink characteristics" ) ; }
	
	// connect component to adapters
	flowNext.setSrcIface( *this ) ;
	flowNext.setDstIface( *this ) ;
	
	// wrap component and continue with next
	this->flows.push_back( FlowBase::Ptr( &flowNext, boost::serialization::null_deleter() ) ) ;
	this->add( std::forward< Args >( argsRest ) ... ) ;
}

// allow the variadic template to set the class's name
template< typename ... Args >
inline void DemandAlloc::add( const std::string nameArg,
			                           Args&& ...  argsRest ) {
	this->name = nameArg ;
	this->add( std::forward< Args >( argsRest ) ... ) ;
}

} } // namespaces "hawaii::flow"
