// Copyright (C) 2013 by FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of libHawaii.
// 
// libHawaii is free software: you can redistribute it and/or modify it under the terms of the GNU General Public 
// License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later 
// version.
// 
// libHawaii is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with libHawaii. If not, see 
// <http://www.gnu.org/licenses/>.


// base component of a graph-like network of processing steps, pipelines etc.
// ==========================================================================

#pragma once

#include "hawaii/flow/iface.h"
#include "hawaii/common/timer.h"
#include "hawaii/common/tracker.h"
#include "hawaii/common/sync.h"
#include "hawaii/common/helpers.h"
#include "hawaii/GPU/nvccFixHeaderAbove.h"
#include <boost/smart_ptr/shared_ptr.hpp>
#include <boost/thread/mutex.hpp>
#include <boost/thread/thread.hpp>
#include "hawaii/GPU/nvccFixHeaderBelow.h"
#include <set>

namespace hawaii {
namespace flow {

// abstract base e.g. for "hawaii::flow::Pipeline" and others
class FlowBase : protected IfacePull,
                 protected IfacePush {
	
	// pointer for classes wrapping flow components
	public:
	typedef boost::shared_ptr< hawaii::flow::FlowBase > Ptr    ;
	typedef std::vector< Ptr >                          PtrVec ;
	
	// run this component, wait for it to finish or directly stop it
	public:
	virtual void run()  = 0 ;
	virtual void wait() = 0 ;
	virtual void stop() = 0 ;
	
	// Is this flow component a source (i.e. produces work items instead of pulling them, e.g. by loading data from 
	// disk) and/or a sink (i.e. saves or displays results instead of passing them on)?
	public:
	virtual bool isSource() const = 0 ;
	virtual bool isSink()   const = 0 ;
	
	// external interfaces
	// user note: You may leave source/destination unset if this flow component is a source/sink.
	// developer note: Use the wrappers "this->{pull,push}(Try)" below instead of directly the interfaces!
	private:
	IfacePull* srcIface ;
	IfacePush* dstIface ;
	public:
	void setSrcIface( hawaii::flow::IfacePull& srcIfaceArg ) { this->srcIface = &srcIfaceArg ; }
	void setDstIface( hawaii::flow::IfacePush& dstIfaceArg ) { this->dstIface = &dstIfaceArg ; }
	
	// connection between external interfaces and internal flow components
	protected:
	uint64_t serialNumberSource ; // counter to be used only if this is a source
	virtual void pullImpl(          ItemBase::Ptr& itemPtr ) override ;
	virtual bool pullTryImpl(       ItemBase::Ptr& itemPtr ) override ;
	virtual void pushImpl(    const ItemBase::Ptr& itemPtr ) override ;
	virtual bool pushTryImpl( const ItemBase::Ptr& itemPtr ) override ;
	
	// print evaluation
	public:
	std::string name ;
	virtual void eval( size_t indentationName = 0,
	                   size_t indentationData = 0 ) const ;
	
	// c'tor with default initialization, reset internal state, copy c'tor
	public:
	FlowBase( const std::string nameArg = "flow component (unnamed)" ) ;
	virtual void reset() ;
	
	// count processed items, measure performance
	// user note: timers, trackers and "latency()" in milliseconds, "throughput(Limit)()" in frames/second
	protected:
	mutable uint64_t counterItems ;
	mutable hawaii::SyncPair< boost::mutex, std::set< boost::thread::id > > trackerTIUUpdated ;
	mutable hawaii::SyncPair< boost::mutex, std::map< uint64_t, hawaii::Timer > > timersLatency          ;
	mutable                                                     hawaii::Timer     timerThroughputInverse ;
	mutable hawaii::Tracker1DExponential trackerLatency,
	                                     trackerThroughputInverse,
	                                     trackerThroughputInverseUnconstrained,
	                                     trackerWaitedOnPull,
	                                     trackerWaitedOnPush,
	                                     trackerRatio                          ;
	public:
	uint64_t items()                   const { return          this->counterItems                          ; }
	double   latency()                 const { return          this->trackerLatency                        ; }
	double   throughput()              const { return 1000.0 / this->trackerThroughputInverse              ; }
	double   throughputUnconstrained() const { return 1000.0 / this->trackerThroughputInverseUnconstrained ; }
	double   waitedOnPull()            const { return          this->trackerWaitedOnPull                   ; }
	double   waitedOnPush()            const { return          this->trackerWaitedOnPush                   ; }
	double   ratio()                   const { return          this->trackerRatio                          ; }
	
	// optionally limit throughput: You may get or set "INFINITY" for unlimited throughput.
	protected:
	hawaii::Timer::Rep timePullDelta,
	                   timePullNext ;
	public:
	virtual double getThroughputLimit() const ;
	virtual void   setThroughputLimit( const double throughputLimit ) ;
	
	// estimate power consumption
	public:
	virtual double power() const = 0 ;
	
	// ensure correct d'tor
	public:
	virtual ~FlowBase() {}
	
} ; // class "FlowBase"

} } // namespaces "hawaii::flow"
