// Copyright (C) 2013 by FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of libHawaii.
// 
// libHawaii is free software: you can redistribute it and/or modify it under the terms of the GNU General Public 
// License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later 
// version.
// 
// libHawaii is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with libHawaii. If not, see 
// <http://www.gnu.org/licenses/>.


// abstract interface for passing work items through flow components
// =================================================================

#pragma once

#include "hawaii/flow/item.h"
#include "hawaii/common/error.h"
#include "hawaii/GPU/nvccFixHeaderAbove.h"
#include <boost/smart_ptr/shared_ptr.hpp>
#include "hawaii/GPU/nvccFixHeaderBelow.h"

namespace hawaii {
namespace flow {

// abstract class for pulling
class IfacePull {

	// implementation to obtain a work item: Non-"Try" must be blocking if necessary, "Try" must return "true" on 
	// success and "false" on failure.
	protected:
	virtual void pullImpl(    ItemBase::Ptr& itemPtr ) = 0 ;
	virtual bool pullTryImpl( ItemBase::Ptr& itemPtr ) = 0 ;
	
	// interface to obtain a work item: includes casts and error checking
	public:
	template< typename ItemUser > void pull(    boost::shared_ptr< ItemUser >& itemUserPtr ) ;
	template< typename ItemUser > bool pullTry( boost::shared_ptr< ItemUser >& itemUserPtr ) ;
	
	// ensure correct d'tor
	public:
	virtual ~IfacePull() {}
} ;

// abstract class for pushing
class IfacePush {
	
	// implementation to pass on a work item: Non-"Try" must be blocking if necessary, "Try" must return "true" on 
	// success and "false" on failure.
	protected:
	virtual void pushImpl(    const ItemBase::Ptr& itemPtr ) = 0 ;
	virtual bool pushTryImpl( const ItemBase::Ptr& itemPtr ) = 0 ;
	
	// interface to pass on a work item: includes casts and error checking
	public:
	template< typename ItemUser > void push(    const boost::shared_ptr< ItemUser >& itemUserPtr ) ;
	template< typename ItemUser > bool pushTry( const boost::shared_ptr< ItemUser >& itemUserPtr ) ;
	
	// ensure correct d'tor
	public:
	virtual ~IfacePush() {}
} ;

// helper to allow ending a flow without a sink processing step
class Terminator : public IfacePush {
	protected:
	virtual void pushImpl(    const ItemBase::Ptr& itemPtr ) override { (void)itemPtr ;               } ;
	virtual bool pushTryImpl( const ItemBase::Ptr& itemPtr ) override { (void)itemPtr ; return true ; } ;
} ;
extern Terminator terminator ;


// "IfacePull": abstract class for pulling
// ---------------------------------------

// interface to obtain a work item: includes casts and error checking
template< typename ItemUser >
inline void IfacePull::pull( boost::shared_ptr< ItemUser >& itemUserPtr ) {
	
	// get base pointer to item
	hawaii::flow::ItemBase::Ptr itemBasePtr ;
	this->pullImpl( itemBasePtr ) ;
	
	// check actual type
	itemUserPtr = boost::dynamic_pointer_cast< ItemUser >( itemBasePtr ) ;
	HAWAII_ERROR_CONDITIONAL( !itemBasePtr, "obtained non-existent source work item"  ) ;
	HAWAII_ERROR_CONDITIONAL( !itemUserPtr, "obtained source work item of wrong type" ) ;
}
template< typename ItemUser >
inline bool IfacePull::pullTry( boost::shared_ptr< ItemUser >& itemUserPtr ) {
	
	// try to get base pointer to item
	hawaii::flow::ItemBase::Ptr itemBasePtr ;
	const bool success = this->pullTryImpl( itemBasePtr ) ;
	
	// check actual type
	if( success ) { itemUserPtr = boost::dynamic_pointer_cast< ItemUser >( itemBasePtr ) ;
	                HAWAII_ERROR_CONDITIONAL( !itemBasePtr, "obtained non-existent source work item"  ) ;
	                HAWAII_ERROR_CONDITIONAL( !itemUserPtr, "obtained source work item of wrong type" ) ; }
	else          { itemUserPtr.reset() ; }
	return success ;
}

// save casts and copies if using base work item type
template<>
inline void IfacePull::pull( ItemBase::Ptr& itemPtr ) {
	this->pullImpl( itemPtr ) ;
	HAWAII_ERROR_CONDITIONAL( !itemPtr, "obtained non-existent source work item" ) ;
}
template<>
inline bool IfacePull::pullTry( ItemBase::Ptr& itemPtr ) {
	const bool success = this->pullTryImpl( itemPtr ) ;
	if( success ) { HAWAII_ERROR_CONDITIONAL( !itemPtr, "obtained non-existent source work item" ) ; }
	else          { itemPtr.reset() ; }
	return success ;
}

// "IfacePush": abstract class for pushing
// ---------------------------------------

// interface to pass on a work item: includes casts and error checking
template< typename ItemUser >
inline void IfacePush::push( const boost::shared_ptr< ItemUser >& itemUserPtr ) {
	
	// check type (at compile-time) and content (at run-time)
	hawaii::flow::ItemBase::Ptr itemBasePtr = boost::static_pointer_cast< ItemBase >( itemUserPtr ) ;
	HAWAII_ERROR_CONDITIONAL( !itemBasePtr, "tried to push non-existent work item" ) ;
	
	// push
	this->pushImpl( itemBasePtr ) ;
}
template< typename ItemUser >
inline bool IfacePush::pushTry( const boost::shared_ptr< ItemUser >& itemUserPtr ) {
	
	// check type (at compile-time) and content (at run-time)
	hawaii::flow::ItemBase::Ptr itemBasePtr = boost::static_pointer_cast< ItemBase >( itemUserPtr ) ;
	HAWAII_ERROR_CONDITIONAL( !itemBasePtr, "tried to push non-existent work item" ) ;
	
	// try to push
	return this->pushTryImpl( itemBasePtr ) ;
}

// save casts and copies if using base work item type
template<>
inline void IfacePush::push( const ItemBase::Ptr& itemPtr ) {
	HAWAII_ERROR_CONDITIONAL( !itemPtr, "tried to push non-existent work item" ) ;
	this->pushImpl( itemPtr ) ;
}
template<>
inline bool IfacePush::pushTry( const ItemBase::Ptr& itemPtr ) {
	HAWAII_ERROR_CONDITIONAL( !itemPtr, "tried to push non-existent work item" ) ;
	return this->pushTryImpl( itemPtr ) ;
}

} } // namespaces "hawaii::flow"
