// Copyright (C) 2013 by FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of libHawaii.
// 
// libHawaii is free software: you can redistribute it and/or modify it under the terms of the GNU General Public 
// License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later 
// version.
// 
// libHawaii is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with libHawaii. If not, see 
// <http://www.gnu.org/licenses/>.


// GPU initialization and shutdown
// ===============================

#include "hawaii/GPU/init.h"
#include "hawaii/GPU/CUDA.h"
#include "hawaii/GPU/pool.h"
#include "hawaii/common/error.h"

namespace hawaii {
namespace GPU {

// initialize all available GPUs with the following preferences:
// - be able to map page-locked host memory into the device address space
// - yield the controlling CPU thread when synchronizing (instead of potentially having it spin)
// - enable peer-to-peer communication with all other GPUs
void init() {
	
	// check compile-time CUDA support
	#ifndef HAWAII_NO_CUDA
		
		// count GPUs (and do nothing if there are none), save caller's active GPU
		int GPUs ; if( cudaSuccess != cudaGetDeviceCount( &GPUs ) ) { GPUs = 0 ; }
		int GPUCall = -1 ; if( GPUs > 0 ) { HAWAII_ERROR_CHECK_CUDA( cudaGetDevice( &GPUCall ) ) ; }
		
		// set flags for above items #1 and #2 on all GPUs
		// developer note: The heuristic of the default setting "cudaDeviceScheduleAuto" does not account for CPU load, 
		//                 and always yielding only costs about 2% GPU performance with the GEMM-DBA sample application.
		const unsigned int flags = cudaDeviceMapHost | cudaDeviceScheduleYield ;
		for( int GPU = 0 ; GPU < GPUs ; ++GPU ) {
			HAWAII_ERROR_CHECK_CUDA( cudaSetDevice( GPU ) ) ;
			HAWAII_ERROR_CHECK_CUDA( cudaSetDeviceFlags( flags ) ) ;
		}
		
		// enable peer-to-peer access between all GPUs supporting it
		for( int GPU = 0 ; GPU < GPUs ; ++GPU ) {
			HAWAII_ERROR_CHECK_CUDA( cudaSetDevice( GPU ) ) ;
			for( int GPUPeer = 0 ; GPUPeer < GPUs ; ++GPUPeer ) {
				if( GPU != GPUPeer ) {
					int supported ; HAWAII_ERROR_CHECK_CUDA( cudaDeviceCanAccessPeer( &supported, GPU, GPUPeer ) ) ;
					if( supported ) {
						HAWAII_ERROR_CHECK_CUDA( cudaDeviceEnablePeerAccess( GPUPeer, 0 ) ) ;
					}
				}
			}
		}
		
		// restore caller's active GPU
		if( GPUCall >= 0 ) { HAWAII_ERROR_CHECK_CUDA( cudaSetDevice( GPUCall ) ) ; }
		
	#endif
}

// clear libHawaii's GPU memory pools
void reset() {
	
	// check compile-time CUDA support
	#ifndef HAWAII_NO_CUDA
		
		// clear global memory pools for all GPUs and allocation sizes
		hawaii::GPU::memoryPool.clear() ;
		
		// count GPUs (and do nothing if there are none), save caller's active GPU
		int GPUs ; if( cudaSuccess != cudaGetDeviceCount( &GPUs ) ) { GPUs = 0 ; }
		int GPUCall = -1 ; if( GPUs > 0 ) { HAWAII_ERROR_CHECK_CUDA( cudaGetDevice( &GPUCall ) ) ; }
		
		// activate and reset all GPUs
		for( int GPU = 0 ; GPU < GPUs ; ++GPU ) {
			cudaSetDevice( GPU ) ;
			cudaDeviceReset() ;
		}
		
		// restore caller's active GPU
		if( GPUCall >= 0 ) { HAWAII_ERROR_CHECK_CUDA( cudaSetDevice( GPUCall ) ) ; }
		
	#endif
}

} } // namespaces "hawaii::GPU"
